#  tests for modin-0.8.3-py37h03978a9_1 (this is a generated file);
print('===== testing package: modin-0.8.3-py37h03978a9_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import modin
import modin.data_management.functions


if __name__ == "__main__":
    import modin.pandas as pd

    print(pd.__version__)
    print(pd.DataFrame([1,2,3]))
#  --- run_test.py (end) ---

print('===== modin-0.8.3-py37h03978a9_1 OK =====');
