# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

from .default import DefaultMethod


class ObjTypeDeterminer:
    def __getattr__(self, key):
        def func(df, *args, **kwargs):
            prop = getattr(df, key)
            if callable(prop):
                return prop(*args, **kwargs)
            else:
                return prop

        return func


class AnyDefault(DefaultMethod):
    @classmethod
    def register(cls, func, obj_type=None, **kwargs):
        if obj_type is None:
            obj_type = ObjTypeDeterminer()

        return cls.call(func, obj_type=obj_type, **kwargs)
